#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import subprocess
import sys
import os
import math
import string
import smtplib
import re
import urllib
import urllib2
import socket
import inspect


__version__ = "$Revision: $"
__author__  = "Jon Tischler, <tischler@aps.anl.gov>" +\
              "Argonne National Laboratory"
__date__    = "$Date: $"
__id__      = "$Id: $"


testingNotify = False

prowlURL = 'https://prowl.weks.net/publicapi/add'
try:	PROWL_ApiKey = urllib.quote_plus(os.environ['PROWL_ApiKey'])	# this goes to local users cell phone
except:	PROWL_ApiKey = urllib.quote_plus('')		# default is no value
# set the PROWL_ApiKey in .bashrc
SMTPserver = 'apsmail.aps.anl.gov'



def scanLogEntry(scanLogFile,scanNumIn):
	"""
	get the entry from the scanLog.txt file, if scanNum<0 get last one
	EXAMPLE::
		>>> scanLogEntry('fileName',5)
		text from the file
	"""

	try:	scanNum = int(scanNumIn)
	except:	scanNum = -1
	if scanNum<0: scanNum = -1

	if (not os.path.isfile(scanLogFile)): return {}
	try:	f = open(scanLogFile,mode='rt')
	except:	return {}
	line = f.readline(200)
	if not checkFileTypeLine(line,'scanLog'): f.close() ; return {}
	#	statinfo = os.stat(scanLogFile)
	#	length = statinfo.st_size

	# find the last scan number, and leave positined there
	buf = f.read(1000)				# only need first two lines
	keyVals = KeyValuesFromBuffer(buf)
	try:	value = keyVals['lastScanPos']
	except:	f.close() ; return {}
	value = int(value)
	if value<0: f.close() ; return {}
	f.seek(value)
	scanLast = hunt_for_next_scanNum(f) # determine the last scan number
	lastScanP = f.tell()				# postion in file of the start of last scan

	# position to start of the scanNum
	if scanNum<0:						# first the case where we want last scan
		scanNum = scanLast
	else:								# hunt for scan number 'scanNum', use a binary search
		f.seek(0)						# goto start of file
		scanFirst = hunt_for_next_scanNum(f)	# first scan number in file
		firstScanP = f.tell()			# postion in file of the start of first scan
		if (scanNum < scanFirst) or (scanNum > scanLast): f.close() ; return {}
		iscan = scanFirst
		iscan0 = -2
		while iscan != scanNum:			# hunt until iscan == scanNum
			pos = int((lastScanP+firstScanP)/2)
			# print 'firstScanP =',firstScanP,'   pos =',pos,'   lastScanP =',lastScanP,'   iscan =',iscan
			f.seek(pos)
			iscan = hunt_for_next_scanNum(f) 	# first scan number in file
			if iscan==iscan0: f.close(); return{'abc': iscan0} # in case we get trapped
			else: iscan0 = iscan ;
			if iscan == -1: iscan = scanLast	# if we ran to EOF, then use biggest number
			if iscan>scanNum: lastScanP = pos	# reset range to lower or upper half
			else: firstScanP = pos

	# accumulate lines into buf & turn into a dictionary
	buf = ''
	line = '$scan	'+str(scanNum)+'\n'	# ensure that it will go through the loop the first time
	while len(line)>2:					# look for blank line between each scan
		buf += line
		line = f.readline(1000)
	f.close()

	keyVals = KeyValuesFromBuffer(buf)	# buf is the contents of one scan
	try:
		X1 = keyVals['X1']
		keyVals['Xsample'] = X1
		del keyVals['X1']
	except:	pass
	try:
		Y1 = keyVals['Y1']
		keyVals['Ysample'] = Y1
		del keyVals['Y1']
	except:	pass
	try:
		Z1 = keyVals['Z1']
		keyVals['Zsample'] = Z1
		del keyVals['Z1']
	except:	pass

	try:
		X2 = keyVals['X2']
		keyVals['Xwire'] = X2
		del keyVals['X2']
	except:	pass
	try:
		Y2 = keyVals['Y2']
		keyVals['Ywire'] = Y2
		del keyVals['Y2']
	except:	pass
	try:
		Z2 = keyVals['Z2']
		keyVals['Zwire'] = Z2
		del keyVals['Z2']
	except:	pass

	# 34ide:pm500:c0:m1   --> Xsample
	# 34ide:pm500:c0:m2   --> Zsample
	# 34ide:pm500:c0:m3   --> Ysample
	# 34ide:focusSample:H --> Hsample
	# 34ide:focusSample:F --> Fsample
	# 34ide:pm500:c0:m1	  --> Zsample
	# 34ide:pm500:c0:m2   --> Xwire
	# 34ide:pm500:c0:m3   --> Ywire
	# 34ide:focusWire:H   --> Hwire
	# 34ide:focusWire:F   --> Fwire
	motPVs = ['34ide:pm500:c0:m1.VAL','34ide:pm500:c0:m2.VAL','34ide:pm500:c0:m3.VAL',\
	'34ide:focusSample:H.VAL','34ide:focusSample:F.VAL','34ide:pm500:c1:m1.VAL','34ide:pm500:c1:m2.VAL',\
	'34ide:pm500:c1:m3.VAL','34ide:focusWire:H.VAL','34ide:focusWire:F.VAL']
	motNames = ['Xsample','Zsample','Ysample','Hsample','Fsample','Zsample',\
	'Xwire','Ywire','Hwire','Fwire']
	for item in keyVals:				# substitute names for PVs
		if keyVals[item] in motPVs:
			i = motPVs.index(keyVals[item])
			keyVals[item] = motNames[i]

	return keyVals

def hunt_for_next_scanNum(f):
	""" Hunt for start of next scan in opened file f starting at the current postion """
	scanNum = -1						# default to bad number
	line = f.readline(200)
	j = 0
	while len(line)>0 and scanNum<0:	# break out at EOF, or if scanNum found
		i = line.find('$scan',0,5)
		if i==0:						# found a possible $scan
			skey = KeyValuesFromBuffer(line)
			try:	scanNum = skey['scan']
			except:	pass
		line = f.readline(1000)
	return int(scanNum)



def checkFileTypeLine(lineIn,typeLIstIn):
	""" typeListIn is either a list of strings or just a string """
	if len(typeLIstIn)<1 or len(lineIn)<2: return False
	line = lineIn
	if (line[0] != '$'): return False			# always have to start with a '$'
	ic = line.find('//')					# remove any comment if present
	if ic>0:	line = line[0:ic]

	line = line.replace('\r','\n')
	i = line.find('\n')
	if (i>0): line = line[0:i]

	line = line.replace('\r',' ')
	line = line.replace('\n',' ')
	line = line.replace('\t',' ')
	line = line.rstrip()					# remove any trailing white space
	while (line.find('  ') >= 0):			# change space runs to single space
		line = line.replace('  ',' ')		# double space to single space
	line = line.replace(',',' ')			# also change commas and semcolons to space
	line = line.replace(';',' ')
	flist = line.split(' ')					# first is tag, others are file types

	if isinstance(typeLIstIn,str):			# if only a string is passed, make it a list
		typeLIst = [typeLIstIn]
	elif isinstance(typeLIstIn,list):
		typeLIst = typeLIstIn
	else:
		return False

	#first check for the old (deprecated) way of doing things with the type in the tag
	tag = flist[0]
	if (tag[1:] in typeLIst): return True		# found it (note, does not include the '$' in the tag)

	# now check for the new way starting with a '$filetype' tag
	if (tag != r'$filetype'): return False		# does not have correct tag
	for type in flist:
		if (type in typeLIst): return True

	return False
#
def checkFileType(filename,typeLIstIn):
	""" check if a file is of a specified type """
	if len(typeLIstIn)<1: return False
	if (not os.path.isfile(filename)): return False
	try:	f = open(filename,mode='rt')
	except:	return False
	line = f.readline(200)
	f.close()
	return checkFileTypeLine(line,typeLIstIn)


def KeyValuesFromBuffer(buf):			# buf is the contents of a tagged file
	""" for a bunch of lines of tagged values, returns a dictionay of tag/value pairs """
	buf = buf.replace('\r\n','\n')
	buf = buf.replace('\n\r','\n')
	buf = buf.replace('\r','\n')
	buf = buf.replace('\t',' ')
	keyList = buf.split('\n')

	keyVals = {}
	for line in keyList:
		# find the key
		line = line.strip()
		if len(line)<1: continue		# skip blank lines
		if line[0] != '$': continue		# all keys must start with a '$'
		i = line.find(' ')
		if i<0:							# just a key, no value
			keyVals[line[1:]] = ''
			continue

		key = line[1:i]
		# remove comment if present, and get value
		ic = line.find('//')
		if ic>0:	value = line[i:ic]
		else:		value = line[i:]
		value = value.strip()

		value = value.strip("'")		# remove quotes, both kinds
		value = value.strip('"')
		value = value.strip("'")
		keyVals[key] = value
#		keyVals[key] = value.strip()

	return keyVals
#
def KeyValuesFromFile(filename):
	""" read a file of tagged values, and convert contents to a dictionary """
	if (not os.path.isfile(filename)): return {}
	try:	f = open(filename,mode='rt')
	except:	return {}
	buf = f.read(200000)
	f.close()

	return KeyValuesFromBuffer(buf)


def splitComment(s,comID):
	"""
	split s into part before comID and part after comID, does not return anything past a new line
	comID is the identifier for the comments.  e.g. comID = '//' returns a tuple with part before 
	comID and part after comID, comID is not returned.
	if s or comID are not strings, then just return s as it is is first part of tuple.
	"""
	if not (type(s) is str): return (s,'')
	i = s.find('\n')							# consider nothing past new line
	if i>=0: s = s[0:i]
	if not (type(comID) is str): return (s,'')	# comID is not string, nothing to strip
	if len(comID)<1: return (s,'')				# empty comID

	i = s.find(comID)							# points to start of comment
	if i<0: return (s,'')						# no comment, so just return everything

	if i>0: s1 = s[0:i]							# part before comment
	else: s1 = ''
	s2 = s[i+len(comID):]						# part after comment
	return (s1,s2)




def compressVertically(s):
	while '\t\n' in s:
		s = s.replace('\t\n','\n')
	while ' \n' in s:
		s = s.replace(' \n','\n')
	while '\n\n' in s:
		s = s.replace('\n\n','\n')
	return s


def indent1Tab(lines):
	""" add 1 tab to the start of each line in lines """
	lines = '\t'+lines
	lines = lines.replace('\n', '\n\t')
	lines = lines.rstrip('\t')
	return lines


def str2bool(v):
	""" returns TRUE or FALSE based on value of string """
	return v.lower() in ['yes', 'true', 't', 'y', '1', '-1']


def any2bool(v):
	""" returns TRUE or FALSE based on value of input, accepts string, numeric, boolean, & None """
	if (type(v) is bool):	return v
	try:	return bool(long(v))	# v is an int, or a long, or a string that looks like one
	except:	pass
	try:	return bool(float(v))	# v is a float or a string that looks like one
	except:	pass
	if (type(v) is str):
		try: return v[0].lower() in ('y', 't')	# first char is y or t
		except: return False
	return False					# not numeric, string, boolean, so return False


def getCPUtype():
	""" returns type of CPU, on my Mac is gives 'x86_64' """
	return os.uname()[4]

def getNcores(queueName):
	""" return the number of cores present, returns a minimum of 1 """
	try:	useQueue = isCluster() and (type(queueName) is str) and len(queueName)>0
	except:	useQueue = False
	if useQueue:
		cmd = ['qstat', '-g', 'ct', '-q', queueName]
		proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, )
		stdout_value, stderr_value = proc.communicate('through stdin to stdout')
		if len(stderr_value)>0:
			print >> sys.stderr, 'ERROR --',repr(stderr_value)
			return 1									# in case of error, assume 1 core
		result = repr(stdout_value)[1:-3]				# trim of leading quote and trailing return quote

		i = result.find(queueName)
		if i<0: return 1
		result = result[i+len(queueName):-1].split()
		try:	Ncores = int(result[3])
		except:	Ncores = 1

	elif os.name is 'posix':							# works for both Linux & Darwin
		try:	Ncores = int(os.sysconf("SC_NPROCESSORS_ONLN"))
		except:	Ncores = 1
	elif os.name is 'nt':								# here for Windows
		try:	Ncores = int(os.environ['NUMBER_OF_PROCESSORS'])
		except:	Ncores = 1
	else:												# unknown platform
		Ncores = 1

	Ncores = max(1,Ncores)
	return Ncores


def isCluster():
	"""
	return True if this is a cluster at the APS, a Sun Grid Engine,
	on that computer, the environment variable SGE_CELL is 'orthros'
	"""
	return os.getenv('SGE_CELL') in ['orthros','blacklab']


def niceDeltaDateTime(dt):
	""" returns a nice string for printing a delta datetime """
	if not(type(dt) is datetime.timedelta): return '?'
	micro = dt.microseconds
	dt = datetime.timedelta(dt.days,dt.seconds)
	if micro > 500000: dt += datetime.timedelta(0,1,0) # round seconds
	return str(dt)


def niceDateTime(dt):
	""" returns a nice string for printing a datetime """
	if not(type(dt) is datetime.datetime): return 'ERROR -- not datetime.datetime type, (is: '+str(type(dt))+')'

	oneDay = datetime.timedelta(1)			# one day
	now = datetime.datetime.now()

	midnight0 = datetime.datetime(now.year,now.month,now.day)	# midnight at start of today
	midnight1 = midnight0 + oneDay
	midnight2 = midnight1 + oneDay
	midnightN1 = midnight0 - oneDay

	if (dt>=midnight0) and (dt<midnight1):	# today
		return dt.strftime('%I:%M:%S%p')
	elif dt>=midnight1 and dt<midnight2:	# tomorrow
		return dt.strftime('Tomorrow, %I:%M:%S%p')
	elif dt>=midnightN1 and dt<midnight0:	# yesterday
		return dt.strftime('Yesterday, %I:%M:%S%p')
	return dt.strftime('%a, %B %d, %Y %I:%M:%S%p')

def getLoginEnvValue(name):
	""" really get environment variables """
	if len(name)<1: return ''
	cmd = '/bin/bash -c "source ~/.bashrc ; /usr/bin/env"'
	proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, )
	stdout_value, stderr_value = proc.communicate('through stdin to stdout')
	if len(stderr_value)>0:
		print >> sys.stderr, 'ERROR --',repr(stderr_value)
		return 1									# in case of error, assume 1 core
	result = repr(stdout_value)[1:-3]				# trim off leading quote and trailing return quote

	find = '\\n'+name+'='
	i0 = result.find(find)
	if i0<0: return ''
	i0 += len(find)									# move to start of value

	i1 = result.find('\\n',i0)
	if i1<i0: i1 = len(result)						# might be at end
	return result[i0:i1]


def fullHostName():
	""" returns the full host name """
	if socket.gethostname().find('.')>=0:
		fullName = socket.gethostname()
	else:
		fullName = socket.gethostbyaddr(socket.gethostname())[1]
		if type(fullName) is list: fullName = str(fullName[0])
	return fullName



def whichCommand(name, flags=os.X_OK):
	"""
	Search PATH for executable files with the given name. Similar to
	the Linux which command.

	On newer versions of MS-Windows, the PATHEXT environment variable will be
	set to the list of file extensions for files considered executable. This
	will normally include things like ".EXE". This fuction will also find files
	with the given name ending with any of these extensions.

	On MS-Windows the only flag that has any meaning is os.F_OK. Any other
	flags will be ignored.

	@type name: C{str}
	@param name: The name for which to search.

	@type flags: C{int}
	@param flags: Arguments to L{os.access}.

	@rtype: C{list}
	@param: A list of the full paths to files found, in the
	order in which they were found.
	"""
	result = []
	exts = filter(None, os.environ.get('PATHEXT', '').split(os.pathsep))
	path = os.environ.get('PATH', None)
	if path is None:
		return []
	for p in os.environ.get('PATH', '').split(os.pathsep):
		p = os.path.join(p, name)
		if os.access(p, flags):
			result.append(p)
		for e in exts:
			pext = p + e
			if os.access(pext, flags):
				result.append(pext)
	return result
	


def str2hkl(s,Nmin=3,Nmax=3):
	"""
	Take the string s and extract the h,k,l as a list
	insists on at least Nmin items, and no more than Nmax items
	This is mainly used to interpret a string containing an hkl which
	is why Nmin and Nmax default to 3.
	raises a TypeError or ValueError on error

	EXAMPLE::
		>>> str2hkl('1 1e-2 3')
		[1.0, 0.01, 3.0]

		>>> str2hkl('1 10 3')
		[1, 10, 3]

		>>> str2hkl('-103')
		[-1, 0, 3]

		>>> str2hkl('002')
		[0, 0, 2]

		>>> str2hkl('0024',Nmax=4)
		[0, 0, 2, 4]
	"""
	try:	s = s.encode()
	except:	raise TypeError('The input s = %r, is not a string' % s)
	s = re.sub('[ \t,;]',' ',s)	# change all separators to a space
	s = s.lower()
	s = s.replace("e+","e")		# "e+" is redundant
	s = s.replace("e-","e_")	# temporarily change "e-" to "e_"
	s = s.replace('-',' -')		# ensure a space separator
	s = s.replace('+',' +')		# ensure a space separator
	s = s.replace("e_","e-")	# change "e_" back to "e-"
	s = s.strip(' {[()]}')		# in case hkl was enclosed in "()", "[]", or "{}"
	while s.find('  ')>=0:
		s = s.replace("  "," ")	# no double spaces

	shkl = s.split(' ')
	if len(shkl)<Nmin:			# need to split some numbers, e.g. '12' -> '1','2'
		shkl2 = []
		for ss in shkl:
			if len(ss)<2 or len(ss)==2 and not (ss[0].isdigit()):
				shkl2.append(ss)
			else:
				i0 = 0
				i1 = 1
				if not ss[0].isdigit(): i1 += 1
				while i1<=len(ss):
#					print 'adding ss[%d:%d] = "%s"' % (i0,i1, ss[i0:i1])
					shkl2.append(ss[i0:i1])
					i0 = i1
					i1 = i0+1
		shkl = shkl2

	if len(shkl)<Nmin or len(shkl)>Nmax:
		raise ValueError('Found %d items in %r, but number of items must be in range [%d, %d]' % (len(shkl),s,Nmin,Nmax))
	hkl = []
	isfloat = False
	for x in shkl:
		try:
			x = float(x)			# yes they must be numbers
			isfloat = isfloat or not x.is_integer()
			hkl.append(x)
		except:
			raise TypeError("Elements must be integer or float, not"+str(type(x)))

	if not isfloat: hkl = [int(i) for i in hkl]
	return hkl


def hkl2str(hklIN,maxMag=1e-14):
	"""
	Format [h,k,l] into a string of acceptable minimal length.
	Return the value or the element that follows after the given value.
	raises a ValueError or ValueError on error

	EXAMPLE::
		>>> hkl2str([1, 2, 3])
		"123"

		>>> hkl2str([1.1, 2, 3])
		"1.1, 2, 3"

		>>> hkl2str([1, 20, 3])
		"1 20 3"
	"""
	if not hasattr(hklIN, '__iter__'):			# this works for list and numpy.array, but fails for strings
		raise TypeError('The input = %r, this is not iterable (the input is ususally a list).' % hklIN)
	try:
		if len(hklIN)<1: raise ValueError('input list is empty')
	except:
		raise TypeError('Could not determine the length of the input list = %r, probably not a list' % hklIN)

	hkl = []
	for x in hklIN:
		try:	x = float(x)
		except:	TypeError('from the input, %r is not a number' % x)
		if abs(x)<maxMag: x = 0
		elif x.is_integer(): x = int(x)
		hkl.append(x)

	allSingle = allInt = True
	for x in hkl:
		allSingle = allSingle and (abs(x)<10) and (type(x) is int)
		allInt = allInt and (type(x) is int)

	out = ''
	for x in hkl:						# assemble the output string
		if allSingle: out += str(x)		# only integers in range [-9,-8,...0,...8,9]
		elif allInt: out += str(x)+' '	# only integers, but some >=10 or <=-10
		else: out += str(x)+', '		# contains floats
	out = out.strip(', ')
	return out



def test_hkl2str():
	test1_hkl2str([1.1,2,3])
	test1_hkl2str([1,20,3])
	test1_hkl2str([1,-20,3])
	test1_hkl2str([2])
	test1_hkl2str([0,0,2])
	test1_hkl2str([1,2,3,4,5,6])
	print ' '
	test1_hkl2str([])				# the following should produce an exception
	test1_hkl2str(None)
	test1_hkl2str('abc')
	test1_hkl2str('1 2 4')
	test1_hkl2str(5.5)

def test_str2hkl():
	test1_str2hkl('1 1e-2 3')
	test1_str2hkl('1+2+3')
	test1_str2hkl('1 10 3')
	test1_str2hkl('-103')
	test1_str2hkl('002')
	test1_str2hkl('0024',Nmax=4)
	print ' '
	test1_str2hkl('0')				# the following should produce an exception
	test1_str2hkl('01234')
	test1_str2hkl(1.2)

def test1_hkl2str(listIn,maxMag=1e-14):
	try:
		out = 'hkl2str(%s) = "%s"' % (repr(listIn),hkl2str(listIn,maxMag=maxMag))
	except Exception, e:
		out = 'hkl2str(%s) FAILED,  %s' % (repr(listIn),repr(e))
	print out

def test1_str2hkl(s,Nmin=3,Nmax=3):
	try:
		out = 'str2hkl(%s) = "%s"' % (repr(s),str2hkl(s,Nmin=Nmin,Nmax=Nmax))
	except Exception, e:
		out = 'hkl2str(%s) FAILED,  %s' % (repr(s),repr(e))
	print out




""" ============================================================================
	============================ Start of statusBuf ============================
"""

class statusBuf(object):

	def __init__(self, bufMax=100, indentStep=2, showRepeats=False):
		"""
		================== ==============================================================
		input parameters    description
		================== ==============================================================
		bufMax              max number of entries in the status buffer
		indentStep          number of spaces to indent each level (default is 2)
		showRepeats         in self.add, optionally do not add if line == the last line
		================== ==============================================================
		""" 
		try:	self.bufMax = int(bufMax)
		except:	raise ValueError('bufMax = %r, it must be an integer' % bufMax)

		try:	indentStep = int(indentStep)
		except:	raise ValueError('indentStep = %r, it must be a positive integer' % indentStep)
		if (indentStep<0): indentStep = 2
		self.indentStep = indentStep				# number of spaces for each indent
		try:	self.showRepeats = bool(showRepeats)
		except:	raise ValueError('showRepeats = %r, it must be a boolean' % showRepeats)
		self.buf = list()							# ring buffer to hold status messages

	def __repr__(self):
		""" Return string representation for statusBuf. """
		return 'statusBuf(bufMax=%r, indentStep=%r, buf=%r)' % (self.bufMax,self.indentStep,self.buf)

	def __str__(self):
		""" Return string value for statusBuf. """
		return repr(self.buf)

	def __len__(self):
		""" Allows use of len(statusBuf) syntax """
		return len(self.buf)

	def len(self):
		""" Allows use of statusBuf.len() syntax """
		return len(self.buf)


	def clear(self):
		""" Clear the status buffer, do not reset bufMax, indentStep, or showRepeats. """
		self.buf = list()


	def add(self,line,out=None):
		"""
		Add to the status buffer, a ring buffer of lenth bufMax.
		The parameter 'line' may be another statusBuf, they can be nested.
		"""
		try:	Nbuf = len(self.buf)
		except:	return 0
		if not line: return Nbuf					# line is empty, nothing to add

		if type(line) is str or type(line) is unicode:
			if line.find('ERROR') >= 0:				# for errors, add file name and line number where addToStatusBuf() was called
				frame = inspect.currentframe()
				try:
					ll = inspect.getouterframes(frame)[1]
					place = '"%s" -- line %g,  ' % (ll[1], ll[2])
					del ll
				except:
					place = ''
				finally:
					del frame
				line += place

		try:	lastLine = self.buf[-1]				# get lastLine to check for repeats
		except:	lastLine = ''
		if self.showRepeats or (lastLine != line):	# skip adding if line is a repeat
			if len(self.buf)>(self.bufMax-1):		# trim off oldest entry
				self.buf = self.buf[1:]
			try:	self.buf.append(line.buf)		# if we are adding another statusBuf
			except:	self.buf.append(line)
			if type(out) is file: print >> out, line # also print the line, but don't print repeats
		return len(self.buf)


	def Print(self, out=None, indentLevel=0):
		""" print the contents of self.buf in a nice way """
		sout = self.buf2str(indentLevel=indentLevel)
		if sout:
			if out:	print >> out, sout
			else:	print sout


	def buf2str(self, buf=None, indentLevel=0):
		"""
		Returns printable string of status buffer. Usually called by Print()
		Do NOT specify buf or indentLevel when you call this. They are only for recursive use.
		"""
		top = buf is None							# used on first entry
		if top: buf = self.buf
		indent = ''.ljust(indentLevel * self.indentStep)	# set indent to number of spaces
		out = ''

		if type(buf) is list:
			for line in buf:
				if type(line) is list: out += self.buf2str(line, indentLevel=indentLevel+1)
				else: out += indent+str(line)+'\n'
		elif line:
			out += indent+str(line)+'\n'

		if top: out = out.rstrip('\n')				# need to remove final LF so print works correct.
		return out

""" ============================= End of statusBuf =============================
	============================================================================
"""



""" ============================================================================
	============================= Start of Notify ==============================
"""

def notify(application, event, msg, addrs, fromaddr=''):
	"""
	============  =================================================================
	Inputs          Desctiption
	============  =================================================================
	application   optional, name of the application sending the notification (str)
	event         optional, name of the event maybe "done processing" (str)
	msg           optional message for the text body (str)
	addrs         where to send the notification use str for one recipient, and tuple or list for multiples
	============  =================================================================
	returns true if an error
	"""
	if type(addrs) is str:		toaddrs = (addrs,)		# change string to tuple
	elif type(addrs) is tuple:	toaddrs = addrs
	elif type(addrs) is list:	toaddrs = addrs
	else:						return True

	# the application that is sending this notification
	try:
		if len(application) < 1: application = inspect.stack()[-1][1]
	except:	application = 'JZTutil.prowl'

	# the event that has occurred
	NoHost = True
	try:
		if len(event) < 1:
			event = str(os.uname()[1])
			NoHost = False
	except:	event = ''

	subject = application
	if len(event) > 1: subject += ', '+event

	if msg is None: msg = ''
	msg = str(msg)
	if len(msg)<1:
		dt = datetime.datetime.now()
		tstr = dt.strftime("%A, %d. %B %Y %I:%M%p")
		if NoHost:	msg = str(os.uname()[1])+' at '+tstr
		else:		msg = 'Notification at: ' + tstr

	errOut = False
	for toaddr1 in toaddrs:
		if validEmail(toaddr1):
			errOut = errOut or sendEmail(fromaddr,toaddr1,subject,msg)

		elif validProwlApiKey(toaddr1):
			errOut = errOut or prowl(toaddr1,application,event,msg,priority='Normal')

	return errOut


def validNotifier(notifyStr):
	""" returns True if notifyStr is valid """
	if not (type(notifyStr) is str): return False
	if validEmail(notifyStr):		return True	# email is valid
	if validProwlApiKey(notifyStr):	return True # prowl apiKey is valid
	return False									# NOT Valid	


def sendEmail(fromaddr,emailAddr,subject,msg):
	""" Send an email """
	global testingNotify, SMTPserver
	if type(emailAddr) is str:		toaddrs = (emailAddr,)		# change string to tuple
	elif type(emailAddr) is tuple:	toaddrs = emailAddr
	elif type(emailAddr) is list:	toaddrs = emailAddr
	else:							return True

	if not validEmail(fromaddr): fromaddr = os.getenv("LOGNAME")+'@'+socket.gethostbyname_ex(socket.gethostname())[0]
	if not validEmail(fromaddr): return True

	if subject is None:	subject = ''
	else:				subject = str(subject)
	if msg is None:		msg = ''
	else:				msg = str(msg)
	if len(subject)<1 and len(msg)<1: return False	# nothing to send, not an error

	for toaddr1 in toaddrs:
		if not validEmail(toaddr1): continue
		if len(subject) > 0:
			msgOut = ("From: %s\r\nTo: %s\r\nSubject: %s\r\n" % (fromaddr,toaddr1,subject)) + msg
		else:
			msgOut = ("From: %s\r\nTo: %s\r\n" % (fromaddr,toaddr1)) + msg

		if testingNotify:
			print '\n'+'about to send Email = \n'+msgOut
			# print "Message of length  %d is:\n%s" % (len(msgOut),msgOut)
		else:
			server = smtplib.SMTP(SMTPserver)
			# server.set_debuglevel(1)
			server.sendmail(fromaddr, toaddr1, msgOut)
			server.quit()

	return False


def validEmail(email_str):
	"""
	returns True for valid email address
	returns False for INvalid emails
	"""
	if not (type(email_str) is str):
		return False
	elif len(email_str) > 6:
		if re.match("^[a-zA-Z0-9._%-+]+@[a-zA-Z0-9._%-]+.[a-zA-Z]{2,6}$", email_str) != None:
			return True
	return False


def prowl(ApiKey,application,event,description,priority='Normal'):
	"""
	send a Prowl notification to local users cell phone
	EXAMPLE::
		>>> JZTutil.prowl("calling application", "event name" ,"optional supporting text"[,priority='Normal'])
		or
		>>> JZTutil.prowl('cellPhone','copy','','')
	
	if the "optional supporting text" is not included then the hostname and time will be sent
	"""
	global prowlURL, PROWL_ApiKey, testingNotify

	if len(PROWL_ApiKey)<1: PROWL_ApiKey = getLoginEnvValue('PROWL_ApiKey')	# set global if not set

	if not validProwlApiKey(ApiKey): return True
	if ApiKey.lower() == 'cellphone': ApiKey = PROWL_ApiKey
	if not validProwlPriority(priority): return True

	try:	application = urllib.quote_plus(str(application))
	except:	application = 'JZTutil.prowl'
	if len(application) < 1: application = 'JZTutil.prowl'
	if len(application) > 255: application = application[0:254]

	try:	event = urllib.quote_plus(str(event))
	except:	event = ''
	if len(event)<1:
		event = str(os.uname()[1])
	if len(event) > 1023: event = event[0:1022]

	if description is None: description = ''
	try:	description = urllib.quote_plus(str(description))
	except:	description = ''
	if len(description) > 9999: description = description[0:9998]

	theFormData = "apikey=" + ApiKey + "&priority=" + priority + "&application=" + application
	if len(event): theFormData += "&event=" + event
	if len(description): theFormData += "&description=" + description
	if testingNotify:
		print '\n'+'about to send prowl = "%s"' % theFormData
	else:
		req = urllib2.Request(url=prowlURL,data=theFormData)
		f = urllib2.urlopen(req)
		# print f.read()
	return False



def validProwlApiKey(ApiKey):
	"""
	return True if Valid api key
	return False if INvalid api key
	"""
	if not( type(ApiKey) is str):							return False	# api key must be a string
	if ApiKey.lower() == 'cellphone':						return True		# special
	elif not (re.search('[^0-9,a-fA-F]',ApiKey) is None):	return False	# api key must be ONLY hex
	elif not( len(ApiKey) == 40 ):							return False	# api key must be 40 bytes long
	return True


def validProwlPriority(priority):
	if type(priority) is int:
		if (priority >= -2) and (priority <=2): return True
	elif type(priority) is str:
		priority = priority.lower()
		priority = priority.replace(' ','')
		if priority == 'verylow': 		return True
		elif priority == 'moderate':	return True
		elif priority == 'normal':		return True
		elif priority == 'high':		return True
		elif priority == 'emergency':	return True
	return False

""" ============================================================================
	============================== End of Notify ===============================
"""


""" ============================================================================
	============================== Start of Units ==============================
"""

def ConvertTemperatureUnits(Tin,unitIN,unitOUT, default='K'):
	"""
	returns Temperature in units of (unitOUT)
	=======		==========================================================
	name        description
	=======		==========================================================
	unitIN		input units, e.g. "C" or "K" or "kT"
	unitOUT		output units same as unitIN
	Tin			input temperature in units of unitIN
	default		default units, used when unitIN or unitOUT are not given
	=======		==========================================================

	Known Temperature units are:
	===========   ============        ================================================================
	   units      abbreviation          description
	===========   ============        ================================================================
	Kelvin          K                  Kelvin, official SI temperature unit
	Celsius         C                  Kelvin = Celsius + 273.15
	Rankine         R                  Kelvin = 1.8 * Rankine
	Fahrenheit      F                  Kelvin = (Fahrenheit+32)/1.8 + 273.15
	Plank           P                  Kelvin = 1.416833e32 * Plank   (also: PlankT, PlankTemperature)
	kT or eV        kT or eV           Kelvin = kT/kB = eV/kB    (kB = Boltzman Const. = 8.6173324e-5)
	===========   ============        ================================================================
	"""

	if unitIN == unitOUT: return Tin				# nothing to do
	if not unitIN: unitIN = default					# unitIN  not given, use default
	if not unitOUT: unitOUT = default				# unitOUT not given, use default

	unitIN = ''.join(unitIN.split())				# removes all whitespace from unitIN
	unitOUT = ''.join(unitOUT.split())				# removes all whitespace from unitOUT

	# second tuple is (factor,offset),   K = (Tin * factor) + offset
	kB = 8.6173324e-5								# Boltzmann constant (eV/K)
	TemperatureUnits = [((u'kelvin',u'k'),(1.0,0.0)), \
		((u'celsius',u'celsiu',u'c'), (1.0, 273.15)), \
		((u'rankine',u'R'), (1.8, 0.0)), \
		((u'fahrenheit',u'F'), (1/1.8, 273.15-32.*5./9.)), \
		((u'plank',u'plankt',u'PlankTemperature',u'P'), (1.416833e32, 0.0)), \
		((u'kbt',u'kt',u'ev'), (1.0/kB, 0.0))]
	"""
		K = C + 273.15
		K = R * (9/5)				# 9/5 = 1.8
		K = (F-32) * (5/9) + 273.15
		K = F * 5/9 + 255.3722222222222		# (273.15- 32*5/9) = 255.3722222222222
		K = Plank * 1.416833e32				# Plank Temperature
		K = eV/kB							# use kT or eV, NOT kT(eV), just one or the other
	"""
	prefixIN,valueIN,offsetIN = find_prefix_unitValue(unitIN,TemperatureUnits, strict=False, IgnoreFinalS=True)
	prefixOUT,valueOUT,offsetOUT = find_prefix_unitValue(unitOUT,TemperatureUnits, strict=False, IgnoreFinalS=True)

	# convert input T(unitIN) to Kelvin

	Tin *= SIprefix2factor(prefixIN)				# apply SI prefix
	K = float(Tin)*valueIN + offsetIN

	# convert K to output T(unitOUT)
	Tout = (K - offsetOUT) / valueOUT
	Tout /= SIprefix2factor(prefixOUT)
	return Tout


def ConvertUnits2meters(unit,defaultLen=None,strict=False):
	"""
	Provides conversion factor from unit to meters, no defaults.  
	If you have a value in units of unit, then
	value * ConvertUnits2meters(unit) will be in meters

	EXAMPLE::
		>>> ConvertUnits2meters('nm')
		1e-9

		usual usage is:
			meters = x_nanometers * ConvertUnits2meters('nm')

	strict refers to the case, if strict is True, then the case must match

	====================        ==========================================================
	length(meter)                 description
	====================        ==========================================================
	1                           meter, metre, m, (defined)
	1e-6                        micron, µm, micrometer, micrometre
	1e-10                       Å, Ang, Angstrom
	1e-12                       fermi
	1.002088e-13                Xunit, Xu
	1.00207697e-13              CuXunit, CuXu
	1.00209952e-13              MoXunit, MoXu
	25.4e-3                     inch, in, inches
	25.4e-6                     mil, (0.001 inch)
	12*25.4e-3                  foot, feet, ft
	3*12*25.4e-3                yard, yd, (3 feet)
	5280*12*25.4e-3             mile, mi, (5280 feet)
	1852                        nauticalmile
	25.4e-3 / 72                point, (1/72 inch)
	12*25.4e-3                  pica, (1/72 foot)
	9.4605284e15                lightyear, ly
	149597870700                au, astronomicalunit
	3.08568025e16               parsec, pc
	0.52917721092e-10           a0, ao, BohrRadius
	1.616199e-35                Plank, PlanckLength
	6*12*25.4e-3                fathom, (6 feet)
	66*12*25.4e-3               chain, (66 feet)
	16.5*12*25.4e-3             rod, (16.5 feet)
	3*5280*12*25.4e-3           league, (3 miles)
	660*12*25.4e-3              furlong, (660 feet)
	0.525                       cubit, (approximate)
	500                         Li, (a Chinese mile, 500 m)
	====================        ==========================================================
	"""
	if not unit: return defaultLen
	elif len(unit)<1: return defaultLen
	elif type(unit) is str: unit = unit.decode('utf-8')	# ensure that unit is unicode

	unit = ''.join(unit.split())			# removes all whitespace from unit
	ipow = unit.find('^')
	power=1									# default power
	if ipow>=0:								# found a power
		spow = unit[ipow+1:]				# string with power
		if spow.find('/') > 0 and spow.find('.')<0: spow += '.'
		power = eval(spow)
		unit = unit[0:ipow]					# string before '^'

	lengthUnits = [(u'Å',1.0e-10), (('micron','micrometer','micrometre'),1.0e-6), \
		(('CuXunit','CuXu'),1.00207697e-13), (('MoXunit','MoXu'),1.00209952e-13), (('Xunit','Xu'),1.002088e-13), ('fermi',1e-12), \
		(('inch','in','inche'),25.4e-3), (('foot','feet','ft'),12*25.4e-3), (('yard','yd'),3*12*25.4e-3), \
		('nauticalmile',1852.0), (('mile','mi'),5280*12*25.4e-3), ('mil',25.4e-6), ('point',25.4e-3 / 72.0), ('pica',12*25.4e-3 / 72.0), \
		(('lightyear','ly'),9.4605284e15), (('au','astronomicalunit'),149597870700.0), (('parsec','pc'),3.08568025e16), \
		(('a0','ao','BohrRadiu'),0.52917721092e-10), (('Plank','PlanckLength'),1.616199e-35), \
		('fathom',6*12*25.4e-3), ('chain',66*12*25.4e-3), ('rod',16.5*12*25.4e-3), ('league',3*5280*12*25.4e-3), ('furlong',660*12*25.4e-3), \
		('cubit',0.525), ('Li',500.0), \
		(('m','meter','metre'),1.0)]

	unit = unit.replace(u'Angstrom',u'Å')	# use Å for Angstrom, no conflicts
	unit = unit.replace(u'Ang',u'Å')
	unit = unit.replace(u'µm','micron')

	prefix, value = find_prefix_unitValue(unit,lengthUnits, strict=strict)

	value *= SIprefix2factor(prefix)		# apply SI prefix
	if not power==1: value = value ** power	# and the power if it is not 1
	return value


def ConvertAngleUnits(angIN,unitIN='None',unitOUT='None',default='deg'):
	""" returns Angle in units of (unitOUT)
	=======		==========================================================
	name        description
	=======		==========================================================
	unitIN		input units
	unitOUT		output units
	angIN		input angle in units of unitIN
	default		default units, used when unitIN or unitOUT are not given
	=======		==========================================================

	Known Angular units are:
	degree, arcsec, arcminute, grad or gradian, cosine(anlge)
	"""

	if unitIN == unitOUT: return angIN				# nothing to do
	if not unitIN: unitIN = default					# unitIN  not given, use default
	if not unitOUT: unitOUT = default				# unitOUT not given, use default

	unitIN = ''.join(unitIN.split())				# removes all whitespace from unitIN
	unitOUT = ''.join(unitOUT.split())				# removes all whitespace from unitOUT

	unitIN = unitIN.replace(u"''",u'sec')			# use 'sec' for double quote
	unitIN = unitIN.replace(u'"',u'sec')
	unitIN = unitIN.replace(u"'",u'min')			# use 'min' for single quote

	unitOUT = unitOUT.replace(u"''",u'sec')			# use 'sec' for double quote
	unitOUT = unitOUT.replace(u'"',u'sec')
	unitOUT = unitOUT.replace(u"'",u'min')			# use 'min' for single quote

	# 400 grad is 360 deg, so conversion is 360/400 = 0.9
	angleUnits = [((u'°',u'degree',u'deg.',u'deg',u'degree'),1.0), \
		(('rad','radian'), 180.0/math.pi), \
		(('arcsec','sec'),1/3600.0), (('arcmin','min'),1/60.0), \
		(('grad','gradian'),0.9), \
		(('cos','cosine','co'),'cos')]				# cosine is special

	prefixIN,valueIN = find_prefix_unitValue(unitIN,angleUnits)
	prefixOUT,valueOUT = find_prefix_unitValue(unitOUT,angleUnits)

	# convert input angle to degree
	angIN *= SIprefix2factor(prefixIN)				# apply SI prefix
	if valueIN == 'cos':							# input is a cosine(angle)
		if angIN >= 1.0: degree = 0.0
		if angIN <= -1.0: degree = 180.0
		else:	degree = math.acos(angIN) * (180.0/math.pi)
	else:
		degree = angIN * valueIN

	# degree is the input angle in degrees
	# convert degree into unitOUT

	if valueOUT == 'cos':							# output is a cosine(angle)
		angOUT = math.cos(degree * math.pi / 180.0)
	else:
		angOUT = degree / valueOUT					# apply units correction

	angOUT /= SIprefix2factor(prefixOUT)			# apply SI prefix
	return angOUT



def find_prefix_unitValue(unitStr,unitDefs, strict=False, IgnoreFinalS=True):
	"""
	from unitStr return a tuplie with (SIprefix,unit convertion factor)
	if strict==True, then unitss name must match exactly, otherwise case insensitive.
	unitStr ignores any enclosing () or []
	Note the Siprefix is always case sensitive.
	"""
	unitStr = unitStr.strip('()[]{}<>')		# ignore enclosing parenthesis, bracket, ...
	unitStr = unitStr.strip(string.punctuation)	# trailing or leading period, comma, ...
	if IgnoreFinalS:
		if unitStr.endswith('s'): unitStr = unitStr[0:-1]	# remove trailing 's'
		if not strict:
			if unitStr.endswith('S'): unitStr = unitStr[0:-1]	# remove trailing 'S' too
	if strict:	unitL = unitStr
	else:		unitL = unitStr.lower()		# lower case verison for finding unit names

	value = prefix = None
	for strs,num in unitDefs:				# num may be a tuple of two or a single number
		if not strict:						# not strict
			if type(strs) is tuple:			# convert all strings in strs to lower case
				strsL = ()
				for s in strs: strsL += (s.lower(),)
			else:
				strsL = strs.lower()		# just a single string
		else:
			strsL = strs

		if unitL.endswith(strsL):
			if type(strs) is str:			# a single string
				prefix = unitStr[:-len(strs)]	# remove units name, leaving SI prefix
				value = num
				break						# found the units, stop searching
			else:							# strs is a tuple
				for s in strsL:
					if unitL.endswith(s):
						prefix = unitStr[:-len(s)]	# remove units name, leaving SI prefix
						value = num
						break				# found the units, stop searching
	if not value:							# no units found
		raise ValueError('find_prefix_unitValue(): Cannot interpret units=%r' % unitStr)

	if type(value) is tuple:
		factor = value[0]
		offset = value[1]
		return (prefix,factor,offset)
	else:
		return (prefix,value)



def SIprefix2factor(prefix):
	"""
	returns number corresponding to the SI prefix.
	Excepth for 'H' or 'K' this is case SENSITIVE.
	====    ====    ======         ====    ====    =============================
	char    num      name          char    num      name 
	====    ====    ======         ====    ====    =============================
	 d      1e-1     deci
	 c      1e-2     centi          h      1e2      hecto (h or H) is acceptable)
	 m      1e-3     milli          k      1e3      kilo (k or K is acceptable)
	 µ      1e-6     micro          M      1e6      Mega
	 n      1e-9     nano           G      1e9      Giga
	 p      1e-12    pico           T      1e12     Tera
	 f      1e-15    femto          P      1e15     Peta
	 a      1e-18    atto           E      1e18     Exa
	 z      1e-21    zepto          Z      1e21     Zeta
	 y      1e-24    yocto          Y      1e24     Yotta
	====    ====    ======         ====    ====    =============================

	EXAMPLE::
		>>> SIprefix2factor('m')
				1e-3
		>>> SIprefix2factor('nK')
				1e-6
	"""
	if not prefix: return 1
	if type(prefix) is str: prefix = prefix.decode('utf-8')	# ensure that prefix is unicode
	prefix = ''.join(prefix.split())		# removes all whitespace from prefix

	keyVals = {'d': 0.1, 'c': 0.01, 'm': 1e-3, u'µ': 1e-6, 'n': 1e-9, 'p': 1e-12, 'f': 1e-15, 'a': 1e-18, 'z': 1e-21, 'y': 1e-24, \
		'h': 100., 'H': 100., 'k': 1e3, 'K': 1e3, 'M': 1e6, 'G': 1e9, 'T': 1e12, 'P': 1e15, 'E': 1e18, 'Z': 1e21, 'Y': 1e24}

	value = 1
	for ch in prefix:
		try:	value *= keyVals[ch]
		except:	raise ValueError('SIprefix2factor(): Unknown SI prefix character "%r".' % ch)

	return value


""" ============================== Test Routines =============================== """


def test_SIprefix(prefix):
	try:
		print '%r  -->  %r' % (prefix, SIprefix2factor(prefix))
	except ValueError as e:
		print 'INVAID -- %r  -->  %r' % (prefix, e)

def test_ConvertUnits2meters(unit,defaultLen=None):
	try:
		if defaultLen is None:
			val = ConvertUnits2meters(unit)
		else:
			val = ConvertUnits2meters(unit, defaultLen=defaultLen)

		if defaultLen is None:	print '%r  -->  %r' % (unit,val)
		else:					print '%r, defaultLen="%r"  -->  %r' % (unit, defaultLen, val)
	except ValueError as e:
		print 'INVAID -- %r, defaultLen="%r"  -->  %r' % (unit, defaultLen, e)

def test_Units():
	if True:
		print ' '
		for prefix in ['µ','m','a','p','M','k','K','h','H','mµ','','q']:
			test_SIprefix(prefix)

	if True:
		print ' '
		for ch in ['m','ft','foot','mm','mm^2','Å',u'Å','Å^2','Ang^2','Å^-1','µm^1/2','µm^0.5','inch']:
			test_ConvertUnits2meters(ch)

		print ' '
		for ch in ['pc','PC','mpc','Mpc','µm^1/2','µm^1/2.','yd','dyd']:
			test_ConvertUnits2meters(ch)

		print ' '
		test_ConvertUnits2meters('',defaultLen=1e-3)
		test_ConvertUnits2meters('xx',defaultLen=1e-3)
		test_ConvertUnits2meters('mxx',defaultLen=1e-3)
		test_ConvertUnits2meters('xm',defaultLen=1e-3)
		test_ConvertUnits2meters('')
		test_ConvertUnits2meters('xx')
		test_ConvertUnits2meters('s')


def test_ConvertAngleUnits(angIN,unitIN,unitOUT):
	try:
		angOUT = ConvertAngleUnits(angIN,unitIN,unitOUT)
		print '%g (%s)  -->  %r (%s)' % (angIN,unitIN,angOUT,unitOUT)
	except ValueError as e:
		print 'INVAID -- %r (%r)"  -->  %r     %r' % (angIN,unitIN, unitOUT, e)

def test_AngleUnits():
	print ' '
	for ch in [u'°','deg','degree','degrees','rad','grad','min','arcmin','sec','arcsec',"'",'"',"''",'mcos','Cosine']:
		test_ConvertAngleUnits(0.5,ch,'degree')

	print ' '
	for ch in [u'°','deg','degree','degrees','rad','grad',"'",'"',"''",'mcos','Cosine']:
		test_ConvertAngleUnits(2,'degree',ch)

	print ' '
	test_ConvertAngleUnits(0.0005,'cos','degree')
	test_ConvertAngleUnits(0.5,'mcos','degree')

	print ' '
	test_ConvertAngleUnits(0.01,'deg','cos')
	test_ConvertAngleUnits(89.,'deg','cos')
	test_ConvertAngleUnits(89.,'deg','mcos')
	test_ConvertAngleUnits(180.,'','rad')

	print ' '
	test_ConvertAngleUnits(.05,'cos','deg')
	test_ConvertAngleUnits(50,'mcos','deg')



def test_ConvertTemperatureUnits(Tin,unitIN,unitOUT):
	try:
		Tout = ConvertTemperatureUnits(Tin,unitIN,unitOUT)
		print '%g (%s)  -->  %r (%s)' % (Tin,unitIN,Tout,unitOUT)
	except ValueError as e:
		print 'INVAID -- %r (%r)"  -->  %r     %r' % (Tin,unitIN, unitOUT, e)

def test_TemperatureUnits():
	print ' '
	test_ConvertTemperatureUnits(1,'C','Celsius')
	test_ConvertTemperatureUnits(1,'K','Kelvin')
	test_ConvertTemperatureUnits(1,'F','fahrenheit')
	test_ConvertTemperatureUnits(1,'R','RANKINE')
	print ' '
	test_ConvertTemperatureUnits(0,'Kelvin','C')
	test_ConvertTemperatureUnits(0,'Celsius','K')
	test_ConvertTemperatureUnits(0,'C','K')
	test_ConvertTemperatureUnits(273.15,'K','C')
	test_ConvertTemperatureUnits(272.15,'K','C')
	test_ConvertTemperatureUnits(1,'Plank','K')
	test_ConvertTemperatureUnits(1,'K','Plank Temperature')
	test_ConvertTemperatureUnits(300,'K','eV')
	test_ConvertTemperatureUnits(300,'K','kT')
	test_ConvertTemperatureUnits(300,'K','meV')
	test_ConvertTemperatureUnits(3e5,'mK','meV')
	test_ConvertTemperatureUnits(300,'K','(eV)')
	test_ConvertTemperatureUnits(300,'K','(meV)')
	print ' '
	test_ConvertTemperatureUnits(32,'F','C')
	test_ConvertTemperatureUnits(-40,'F','C')
	test_ConvertTemperatureUnits(-40,'c','F')
	test_ConvertTemperatureUnits(0,'C','F')
	print ' '
	test_ConvertTemperatureUnits(1,'Kelvin','K')
	test_ConvertTemperatureUnits(1,'Kelvin','C')
	print ' '
	print 'milli'
	test_ConvertTemperatureUnits(1,'mK','K')
	test_ConvertTemperatureUnits(1,'K','mK')
	print ' '
	print ' '
	test_ConvertTemperatureUnits(300,'K','kT(eV)')

""" =============================== End of Units ===============================
	============================================================================
"""


if __name__ == '__main__':
	"""
	Main function for JZTutil.py.
	Test cases for units conversion to verify correct behavior.
	"""
	try:	testGroup = int(sys.argv[1])
	except:	raise ValueError('./srange testGroup,  testGroup must be a bit mask, a combo of 1-2048, use -1 for all')
	print 'showing testGroup = %r' % testGroup

	if testGroup & 1:
		print '\ntestGroup 1'
		test_Units()
		#	./JZTutil.py 1 > UnitTest.txt

	if testGroup & 2:
		print '\ntestGroup 2'
		test_AngleUnits()
		#	./JZTutil.py 2 > AngleTest.txt

	if testGroup & 4:
		print '\ntestGroup 4'
		test_TemperatureUnits()
		#	./JZTutil.py 4 > TemperatureTest.txt

	if testGroup & 8:
		print '\ntestGroup 8'
		test_hkl2str()
		#	./JZTutil.py 8 > hkl2strTest.txt

	if testGroup & 16:
		print '\ntestGroup 16'
		test_str2hkl()
		#	./JZTutil.py 16 > str2hklTest.txt
